#!/usr/bin/env python
#
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2019 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------

import argparse
import datetime
import logging
import os
import re
import string


logger = logging.getLogger(__name__)

__arm_copyright = """/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT """+str(datetime.datetime.now().year)+""" ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */"""


def replace_multiple(substitutions, text):
    # Create a regular expression from all of the dictionary keys
    re_string = r'\b(?:' + r'|'.join(map(re.escape, substitutions.keys())) + r')\b'
    regex = re.compile(re_string)

    # For each match, look up the corresponding value in the dictionary
    return regex.sub(lambda m: substitutions[m.group(0)], text)


def c_eval(expr):
    # Evaluate expression as C. Replace keywords with equivalent
    # python, and evaluate as Python expression. This simple
    # implementation may get precedence wrong.

    expr = expr.replace('!', ' not ')
    expr = expr.replace('&&', ' and ')
    expr = expr.replace('||', ' or ')

    return eval(expr)


class NoopImplementation(object):
    # Stub implementation that doesn't do anything at runtime. Returns errors.
    init_tmpl ="""{copyright}
#include <stddef.h>
{includes}

/* Generated by """+os.path.basename(__file__)+""" */
"""

    fn_tmpl = """
{prototype}
{{
    return {error};
}}
"""

    void_fn_tmpl = """
{prototype}
{{
    return;
}}
"""

    union_fn_tmpl = """
{prototype}
{{
    {rtype} r;
    r.{rfield} = {error};
    return r;
}}
"""


class TrampolineImplementation(object):
    # Call dlopen on the target on startup, and then call dlsym for
    # each symbol on its first call.
    init_tmpl ="""{copyright}
#include <stdio.h>
#include <dlfcn.h>
{includes}

/* Generated by """+os.path.basename(__file__)+""" */

static void *tgt_lib;
static __attribute__((constructor)) void init(void)
{{
    tgt_lib = dlopen("{target}", RTLD_LAZY);
    if (tgt_lib == NULL)
    {{
        fprintf(stderr, "Error: Unable to open %s\\n", "{target}");
    }}
}}

static __attribute__((destructor)) void term(void)
{{
    if (tgt_lib)
    {{
        dlclose(tgt_lib);
        tgt_lib = NULL;
    }}
}}
"""

    fn_tmpl = """
{prototype}
{{
    static {rtype} (*fn)({arg_types}) = NULL;
    if (fn != NULL)
    {{
        return fn({args});
    }}
    else
    {{
        *(void **)(&fn) = dlsym(tgt_lib, "{symbol}");
        if (fn != NULL)
        {{
            return fn({args});
        }}
        else
        {{
            fprintf(stderr, "%s not found\\n", "{symbol}");
        }}
    }}

    return {error};
}}
"""

    void_fn_tmpl = """
{prototype}
{{
    static void (*fn)({arg_types}) = NULL;
    if (fn != NULL)
    {{
        fn({args});
    }}
    else
    {{
        *(void **)(&fn) = dlsym(tgt_lib, "{symbol}");
        if (fn != NULL)
        {{
            fn({args});
        }}
        else
        {{
            fprintf(stderr, "%s not found\\n", "{symbol}");
        }}
    }}

    return;
}}
"""

    union_fn_tmpl = """
{prototype}
{{
    static {rtype} (*fn)({arg_types}) = NULL;
    if (fn != NULL)
    {{
        return fn({args});
    }}
    else
    {{
        *(void **)(&fn) = dlsym(tgt_lib, "{symbol}");
        if (fn != NULL)
        {{
            return fn({args});
        }}
        else
        {{
            fprintf(stderr, "%s not found\\n", "{symbol}");
        }}
    }}

    {rtype} r;
    r.{rfield} = {error};
    return r;
}}
"""


class CCommentStateMachine(object):
    ST_NONE=0
    ST_MULTILINE=1
    comment_re = re.compile(r"//.*")
    single_line_comment_re = re.compile(r"/\*.*\*/")
    start_multi_line_comment_re = re.compile(r"/\*.*$")
    end_multi_line_comment_re = re.compile(r"^.*\*/")

    def __init__(self):
        self.state = self.ST_NONE

    def process_comments(self, args, line):
        # return line with comments stripped
        # TODO: if we're interested in comment content add a call to a handler function

        if self.state == self.ST_MULTILINE:
            (l, c) = self.end_multi_line_comment_re.subn('', line)
            if c == 1:
                self.state = self.ST_NONE
                line = l
            else:
                line = ""
        else:
            # Strip single line comments
            line = self.single_line_comment_re.sub('', line)

            # Strip the start of multiline comments
            (line, c) = self.start_multi_line_comment_re.subn('', line)
            if c == 1:
                self.state = self.ST_MULTILINE

            # Strip c++ comments
            line = self.comment_re.sub('', line)

        return line


class CPreprocessorStateMachine(object):
    # Virtual class that identifies and strips preprocessor
    # directives. A concrete class must implement handle_preproc() in
    # order to actually handle preprocessor directives.
    ST_NONE=0
    ST_ACCUM=1
    start_pp_re = re.compile(r"^\s*#")
    complete_pp_re = re.compile(r"^#.*[^\\]$",
                                re.DOTALL)


    def __init__(self):
        self.state = self.ST_NONE
        self.text = ""
        self.excluding = False


    def process_pp(self, args, line, exclude):
        # Strip preprocessing directives.
        #
        # If exclude is true, also process #ifs to exclude code.
        #
        # Pass preprocessor directive to virtual function
        # handle_preproc to take care of.

        if self.state == self.ST_ACCUM:
            self.text += line
        else:
            self.text = line

        m = self.complete_pp_re.match(self.text.strip())
        if m:
            # Handle preprocessor directives.
            logger.info("preproc")
            self.excluding = self.handle_preproc(args, m.group(0))

            # A preprocessor statement will have consumed all the text, so reset
            line = ""
            self.state = self.ST_NONE
            self.text = ""

        elif self.state != self.ST_ACCUM:
            # check if we have the start of a preprocessor statement
            self.start_pp_re.match(self.text)
            if m:
                self.state = self.ST_ACCUM

        if self.state == self.ST_ACCUM:
            line = ""

        if exclude and self.excluding:
            # We don't want to process text protected by preprocessor conditionals
            line = ""

        return line


class PPStateMachine(CPreprocessorStateMachine):
    preproc_cond_re = re.compile(r"#([a-z]+)")
    if_re = re.compile(r"#(?:if|ifdef|ifndef)\s+(.*)")
    elif_re = re.compile(r"#elif\s+(.*)")
    define_re = re.compile(r"#define\s+([A-Za-z_][A-Za-z0-9_]*)(\s+(.*))?")
    identifier_re = re.compile(r'\b([A-Za-z_][A-Za-z0-9_]*)\b')
    defined_re = re.compile(r'\bdefined\s*\(\s*([A-Za-z_][A-Za-z0-9_]*)\s*\)')
    defined2_re = re.compile(r'\bdefined\s+([A-Za-z_][A-Za-z0-9_]*)\b')


    def __init__(self, defines):
        super(PPStateMachine, self).__init__()
        self.defines = defines.copy()
        self.if_depth = 0
        self.state_stack = [True]
        self.seen_true_stack = [True]


    def handle_preproc(self, args, text):

        # Evaluate preprocessor conditionals
        #
        # return True if we think preprocessor conditionals indicate
        # that lines following this should not be compiled.

        m = self.preproc_cond_re.match(text)
        if m:
            directive = m.group(1)
            if directive in ('if', 'ifdef', 'ifndef', 'else', 'elif', 'endif'):
                if directive in ('if', 'ifdef', 'ifndef'):
                    current_state = self.state_stack[self.if_depth]
                    self.if_depth += 1

                    m = self.if_re.match(text)
                    if m:
                        expr = m.group(1)
                        result = False
                        if directive == 'ifdef':
                            result = expr in self.defines

                        elif directive == 'ifndef':
                            result = not(expr in self.defines)

                        else:
                            result = (self.pp_eval(expr) != 0)
                        logger.info("{} {} is {}, depth {}".format(directive, expr, result, self.if_depth))
                        self.state_stack.append(result and current_state)
                        self.seen_true_stack.append(result)

                elif directive == 'endif':
                    logger.info("{} depth {}".format(directive, self.if_depth))
                    self.if_depth -= 1
                    self.state_stack.pop()
                    self.seen_true_stack.pop()

                elif directive == 'elif':
                    if self.seen_true_stack[self.if_depth]:
                        self.state_stack[self.if_depth] = False
                        logger.info("{} already taken, depth {}".format(directive, self.if_depth))

                    else:
                        m = self.elif_re.match(text)
                        if m:
                            expr = m.group(1)
                            result = (self.pp_eval(expr) != 0)
                            logger.info("{} {} is {}, depth {}".format(directive, expr, result, self.if_depth))
                            prev_stack_state = self.state_stack[self.if_depth-1]
                            self.state_stack[self.if_depth] = result and prev_stack_state
                            self.seen_true_stack[self.if_depth] = result

                elif directive == 'else':
                    if not self.seen_true_stack[self.if_depth]:
                        self.state_stack[self.if_depth] = True
                        self.seen_true_stack[self.if_depth] = True
                        logger.info("{} not previously taken, depth {}".format(directive, self.if_depth))
                    else:
                        logger.info("{} already taken, depth {}".format(directive, self.if_depth))

            elif directive == 'define':
                # Accumulate simple macro definitions
                # Not expected to handle macro functions
                if self.state_stack[self.if_depth]:
                    m = self.define_re.match(text)
                    if m:
                        ident = m.group(1)
                        value = m.group(3)
                        self.defines[ident] = value
                    else:
                        logger.warning("Unhandled #define: {}".format(text))

        return not self.state_stack[self.if_depth]


    def expand_macros(self, text, ignore=[]):
        # For each word in text, replace it with its macro expansion if present.
        #
        # The replacement needs to be recursive, __except__ where the
        # macro expansion results in itself (or a macro already expanded)
        expansions = {}

        # First identify words in text and find the macro expansions
        words = self.identifier_re.finditer(text)
        for mword in words:
            word = mword.group(1)
            if word in self.defines:
                if word not in ignore:
                    if word not in expansions:
                        # recursively expand the macro
                        expansion = self.defines[word]
                        new_ignore = [word]
                        new_ignore.extend(ignore)
                        expansion = self.expand_macros(expansion, new_ignore)
                        expansions[word] = expansion
                    #else we already have this expansion

        if len(expansions) > 0:
            # Do all the replacements at once.
            #
            # This should avoid the problem where one macro expansion
            # results in the name of another macro which shouldn't be
            # further expanded (as the recursion was limited)
            return replace_multiple(expansions, text)
        else:
            return text


    def pp_eval(self, expr):
        # Evaluate a preprocessor expression
        # Expand 'defined(x)' and 'defined x'
        # TODO: Handle concatenation and stringification ## and #

        # For each match, look up the corresponding value in the dictionary
        # TODO: can we combine the REs?
        expr = self.defined_re.sub(lambda m: self.defines.get(m.group(0), '0'), expr)
        expr = self.defined2_re.sub(lambda m: self.defines.get(m.group(0), '0'), expr)
        expr = self.expand_macros(expr)
        return c_eval(expr)


arglist_re = re.compile(r"""\((.*)\)\s*           # Parenthesized argument list
                            [A-Za-z0-9_]*\s*$""", # optional decorator
                        re.DOTALL | re.VERBOSE)
arg_identifier_re = re.compile(r"""^(                      # Type group
                                     (?:(?:const|restrict|struct|enum|union|signed|unsigned)\s+)*
                                                           # Qualifiers
                                     [A-Za-z0-9_]+\b\s*      # Typename
                                     (?:(?:(?:const|restrict)|\*)\s*)* # Pointer qualifiers
                                    )
                                    ([A-Za-z0-9_]+)?        # Identifier
                                    (\[[A-Za-z0-9_\s]*\])? # Array subscript
                                 """,
                                re.VERBOSE)
# Assume function pointer args do not have function pointer args or returns
fnarg_identifier_re = re.compile(r"""^([A-Za-z0-9_\*]+\s*           # Function return
                                       \(\s*[A-Za-z0-9_]*\s*\*)\s*  # Decorator and pointer
                                     ([A-Za-z0-9_]+)\s*             # Identifier
                                     (\)\s*
                                      \([^)]*\))                    # Argument list
                                  """,
                                re.DOTALL | re.VERBOSE)
# Use to recover to the next argument
arg_recover_re = re.compile(r"^[^,]+")
fn_identifier_re = re.compile(r"([A-Za-z0-9_]+)$")
void_re = re.compile(r"\bvoid\b")
voidp_re = re.compile(r"\bvoid\s*\*")
union_re = re.compile(r"^union\s+([A-Za-z0-9_]+)$")


def handle_export(args, text):
    prototype = text[:text.rfind(')')+1].strip() # take off everything after the last ')'
    symbol = ""
    arg_values = ""
    arg_types = ""
    rtype = ""

    rewrite_prototype = False
    arg_typelist = []
    arg_namelist = []

    m = arglist_re.search(prototype)
    if m:
        rem_arg_list = m.group(1).strip('\v\n\t ')
        while len(rem_arg_list) > 0:
            # Identify the next argument.
            m = fnarg_identifier_re.search(rem_arg_list)
            if m:
                typ = m.group(1) + m.group(3)
                arg_typelist.append(typ)
                arg_namelist.append(m.group(2))
            else:
                m = arg_identifier_re.search(rem_arg_list)
                if m:
                    typ = m.group(1).strip()
                    name = m.group(2)
                    if m.group(3):
                        # There's an array subscript, so add a pointer
                        typ += "*"
                    arg_typelist.append(typ)
                    if name:
                        arg_namelist.append(name)
                    elif typ != 'void':
                        arg_namelist.append('arg{}'.format(len(arg_namelist)))
                        rewrite_prototype = True

                else:
                    logger.warning("Unrecognized argument {}".format(rem_arg_list))
                    m = arg_recover_re.search(rem_arg_list)

            rem_arg_list = rem_arg_list[len(m.group(0)):]
            rem_arg_list = rem_arg_list.strip('\v\n\t ,')

        arg_values = ', '.join(arg_namelist)
        arg_types = ', '.join(arg_typelist)

    remainder = arglist_re.sub('', prototype).strip()
    m = fn_identifier_re.search(remainder)
    if m:
        symbol = m.group(1)
    else:
        # No function identifier
        # Check if this is a function pointer decl
        if 'typedef' not in remainder:
            logger.error("No function identifier in {}".format(text))
        else:
            return

    if rewrite_prototype:
        decls = []
        for (i, t) in enumerate(arg_typelist):
            decls.append("{} {}".format(t, arg_namelist[i]))

        # Put the argument list back on with argument names
        prototype = '{}({})'.format(remainder, ", ".join(decls))

    remainder = fn_identifier_re.sub('', remainder).strip()
    remainder = args.decorator_re.sub('', remainder)
    m = void_re.search(remainder)
    m2 = voidp_re.search(remainder)
    m3 = union_re.search(remainder)
    if m and not m2:
        args.outf.write(args.implementation.
                        void_fn_tmpl.format(prototype=prototype,
                                            symbol=symbol,
                                            args=arg_values,
                                            arg_types=arg_types))
    elif m3:
        rtype = remainder.strip();
        ident = m3.group(1)
        (rfield, error) = args.union_error_map.get(ident, "value:0").split(':', 1)
        args.outf.write(args.implementation.
                        union_fn_tmpl.format(prototype=prototype,
                                             symbol=symbol,
                                             args=arg_values,
                                             arg_types=arg_types,
                                             rtype=rtype,
                                             rfield=rfield,
                                             error=error))
    else:
        rtype = remainder.strip();
        if rtype[-1:] == "*":
            error = "NULL"
        else:
            error = args.error_map.get(rtype, 0)
        args.outf.write(args.implementation.
                        fn_tmpl.format(prototype=prototype,
                                       symbol=symbol,
                                       args=arg_values,
                                       arg_types=arg_types,
                                       rtype=rtype,
                                       error=error))


C_ST_GLOBAL=0
C_ST_TYPE=16


def process_file(args, f):

    # Go through the file looking for external declarations.
    #
    # Since a declaration may be multiline we need to keep a number of
    # lines of text to identify is we have a declaration.
    #
    # We can discard lines when we identify they are complete
    # statements that are not an external declaration.

    # Regular expression to look for something that looks like a
    # function declaration that includes the export declarator.
    # This needs to be applied multiline
    if args.export_decorators:
        export_re = re.compile(r"^.*\b"+args.export_decorators+ # export decorator
                               r"""\b.*                         # return type and more decorators
                                   [A-Za-z0-9_]+\s*             # function identifier
                                   \(.*\)                       # argument list
                                   .*;""",                      # more decorators
                               re.DOTALL | re.VERBOSE)
    else:
        # No export decorators. Do our best to match any function declaration
        export_re = re.compile(r"""^(?:(?:[A-Za-z0-9_]+(?:\s|\*)+)*) # return type
                                   (?:\(\s*\*\s*[A-Za-z0-9_]+\s*\)| # function pointer identifier or
                                      [A-Za-z0-9_]+)\s*         # identifier
                                   \(.*\)                       # argument list
                                   .*;""",                      # more decorators
                               re.DOTALL | re.VERBOSE)

    # Used to manage scope
    externc_re = re.compile(r'^\s*extern\s+"C"\s*\{[^}]*$',
                            re.DOTALL)
    scope_re = re.compile(r"{[^}]*$")
    end_scope_re = re.compile(r"}")

    # Used to speed up handling with unions, structs and enums
    start_type_re = re.compile(r""".*(?:enum|struct|union)\s+  # keywords
                                   [A-Za-z_][A-Za-z0-9_]*\s*   # identifier
                                   {[^}]*$""",                 # scope
                               re.VERBOSE)

    # A complete statement
    statement_re = re.compile(r"^.*;",
                              re.DOTALL)

    text = ""
    state = C_ST_GLOBAL
    extern_scope = 0
    scope = 0
    type_depth = 0
    lineno = 0
    csm = CCommentStateMachine()
    ppsm = PPStateMachine(args.defines)
    for line in iter(f.splitlines()):
        lineno += 1

        line = csm.process_comments(args, line)
        line = ppsm.process_pp(args, line, True)

        if state == C_ST_GLOBAL:

            # Collect for the current statement
            if text == "":
                text = line
            else:
                text = text + "\n" + line

            logger.info(":" + text + "#")

            # Look for type declarations for structs and enums that we can skip over
            m = start_type_re.match(text)
            if m:
                logger.info("type++")
                state = C_ST_TYPE
                type_depth += 1
                text = ""
                continue

            # Look for exports,
            m = export_re.match(text)
            if m:
                logger.info("export")
                handle_export(args, m.group(0))

                # The statement may not have consumed the full
                # line, so trim the statement from the text.
                text = text[len(m.group(0)):]
                continue

            # TODO: look for exported variables and warn

            # Ignore any other statements.

            m = externc_re.match(text)
            if m:
                logger.info("extern, scope++: %s", text)
                extern_scope += 1
                text = text[len(m.group(0)):]

            m = scope_re.match(text)
            if m:
                logger.info("scope++: %s", text)
                scope += 1

            m = end_scope_re.match(text)
            if m:
                logger.info("scope--: %s", text)
                if scope > 0:
                    scope -= 1
                elif extern_scope > 0:
                    extern_scope -= 1
                else:
                    logger.error("Scope error on line %d", lineno)
                # Scope terminated
                text = ""

            m = statement_re.match(text)
            if m:
                logger.info("statement")
                text = text[len(m.group(0)):]

        elif state == C_ST_TYPE:
            # We're in a type declaration.
            # Note we do not expect to see type initialization.

            m = scope_re.match(line)
            if m:
                logger.info("type++")
                type_depth += 1

            m = end_scope_re.match(line)
            if m:
                logger.info("type--")
                type_depth -= 1

            if type_depth == 0:
                # Scope terminated
                state = C_ST_GLOBAL

        text = text.strip()


def write_implementation(f, copyright, impl, input, include_path, target):
    # Write the forwarding stub implementation

    includes = []
    for header in input:
        if include_path:
            header = re.sub(r'^.*'+include_path+'/', '', header)
        includes.append('#include "{}"'.format(header))

    f.write(impl.init_tmpl.format(copyright=copyright,
                                  includes='\n'.join(includes),
                                  target=target))


class AppendKeyValuePair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        dest = getattr(namespace, self.dest)
        kv = values.split("=", 1)
        if len(kv) == 2:
            dest[kv[0]] = kv[1]
        else:
            dest[kv[0]] = None


def main():
    logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)

    parser = argparse.ArgumentParser()
    parser.add_argument("input", nargs="+", help="Input header file(s)")
    parser.add_argument("-o", "--output",
                        help="Output C file")
    parser.add_argument("--include-path",
                        help="Prefix to remove from header paths for include directives")
    parser.add_argument("--target", default=None,
                        help="Target library")
    parser.add_argument("--export-decorator", action="append",
                        help="Decorator used to indicate an exported function")
    parser.add_argument("--ignore-decorator", action="append", default=['extern'],
                        help="Decorators to ignore on return types. Note: avoid ignoring decorators that affect type compatibility")
    parser.add_argument("-D", dest="defines", action=AppendKeyValuePair, default={}, metavar="MACRO=VALUE",
                        help="Define macro")
    parser.add_argument("-m", "--error-map", action=AppendKeyValuePair, default={}, metavar="TYPE=ERROR",
                        help="Use error value for return type")
    parser.add_argument("--union-error-map", action=AppendKeyValuePair, default={}, metavar="IDENT=FIELD:ERROR",
                        help="Use error value for return type")
    parser.add_argument("-t", "--trampoline", action="store_true", default=False,
                        help="Generate trampolines")
    parser.add_argument("--copyright", action="store_true", default=False,
                        help="Include ARM copyright text")

    args = parser.parse_args()

    if args.trampoline and args.target is None:
        parser.error("--trampoline requires --target to be specified")
    elif args.trampoline:
        args.implementation = TrampolineImplementation
    else:
        args.implementation = NoopImplementation

    # Setup a regular expression which will catch all decorators as a word
    decorators = []
    args.export_decorators = ''
    if args.export_decorator:
        decorators.extend(args.export_decorator)
        args.export_decorators = "(?:{})".format('|'.join(args.export_decorator))
    if args.ignore_decorator:
        decorators.extend(args.ignore_decorator)
    re_string = r'\b(' + r'|'.join(decorators) + r')\b'
    args.decorator_re = re.compile(re_string)

    with open(args.output, 'w') as outf:
        args.outf = outf
        write_implementation(args.outf, __arm_copyright if args.copyright else '',
                             args.implementation,
                             args.input, args.include_path, args.target)

        # Then write the stubs for everything exported by each header
        for inf in args.input:
            with open(inf, 'rb') as f:
                process_file(args, f.read().decode('UTF-8'))


if __name__ == "__main__":
    main()
