#!/bin/bash
set -e

usage() { echo "Usage: $0 [COMMAND make/clean/install/sign] [-p profile_name] {-b debug}"; exit; }

COMMAND=$1
if [ -z "$COMMAND" ]; then
    usage
fi

shift

while getopts "p:b:?" argv
do
    case "$argv" in
        p)
            PROFILE_NAME=${OPTARG} ;;
        b)
            BUILD_MODE=${OPTARG} ;;
        ?)
            usage; exit; ;;
    esac
done

#### check necessary parameters #####
if [ -z $PROFILE_NAME ]; then
    echo "select a profile in ./configs/ with -p (e.g. -p kernel_only.cfg)"
    usage
fi

if [ -z $BUILD_MODE ]
then
    BUILD_MODE="release"
elif [ $BUILD_MODE != "release" ] && [ $BUILD_MODE != "debug" ]
then
    echo "invalid build mode follows with -b"
    exit
fi


##### get local path and read config #####
DDK_KM_DIR=`pwd`
PRJ_SYSTEM_PATH=$DDK_KM_DIR/../../../../../../system
cd $PRJ_SYSTEM_PATH
source "$PRJ_SYSTEM_PATH/configs/$PROFILE_NAME"
cd $DDK_KM_DIR


##### check needed parameters
if [ -z "$ARCH" ]; then
    ARCH=arm
fi
if [ -z "$KERNEL_PATH" ]; then
    echo "KERNEL_PATH does not set in $PROFILE_NAME"
    exit
fi
##### set parameters for building #####
export ARCH=arm
export KDIR=$KERNEL_PATH
export CROSS_COMPILE=$PRJ_SYSTEM_PATH/tmp/toolchain/$K_TOOLCHAIN_NAME/bin/asdk-linux-
export KERNEL_TOOLCHAIN_DIR=$PRJ_SYSTEM_PATH/tmp/toolchain/$K_TOOLCHAIN_NAME/
export KERNEL_TOOLCHAIN_TARBALL=$PRJ_SYSTEM_PATH/toolchain/gcc5/$K_TOOLCHAIN_NAME.tar.bz2
export CC=$PRJ_SYSTEM_PATH/tmp/toolchain/$K_TOOLCHAIN_NAME/bin/asdk-linux-gcc
export CXX=$PRJ_SYSTEM_PATH/tmp/toolchain/$K_TOOLCHAIN_NAME/bin/asdk-linux-g++

if [[ $K_TOOLCHAIN_NAME = *llvm* ]]; then
    export CROSS_COMPILE=$PRJ_SYSTEM_PATH/tmp/toolchain/$K_TOOLCHAIN_NAME/bin/arm-linux-musleabi-
    export CC=clang
    export CXX=clang++
fi

export MALI_PLATFORM=$PLATFORM_ID
echo "MALI_PLATFORM="$PLATFORM_ID

##### check toolchain #####
if [ -z $KERNEL_TOOLCHAIN_CHECK_FILE ]
then
        if [ "$ARCH" = "arm64" ]
        then
                export KERNEL_TOOLCHAIN_CHECK_FILE=$KERNEL_TOOLCHAIN_DIR/bin/aarch64-linux-ld.bfd
                echo $KERNEL_TOOLCHAIN_CHECK_FILE
                echo "use arm 64 toolchain checkfile"
        else
                export KERNEL_TOOLCHAIN_CHECK_FILE=$KERNEL_TOOLCHAIN_DIR/bin/arm-linux-gnueabi-ld.bfd
                echo "arm 32 mode 0 checkfile"
        fi
fi

if [[ $K_TOOLCHAIN_NAME = *llvm* ]]; then
    export KERNEL_TOOLCHAIN_CHECK_FILE=$KERNEL_TOOLCHAIN_DIR/bin/arm-linux-musleabi-ld.bfd
fi

if [ ! -f $KERNEL_TOOLCHAIN_CHECK_FILE ]
then
    if [ ! -d "$PRJ_SYSTEM_PATH/tmp/toolchain" ]
    then
        mkdir -p $PRJ_SYSTEM_PATH/tmp/toolchain
    fi

    cd $PRJ_SYSTEM_PATH/tmp/toolchain
    rm -rf $KERNEL_TOOLCHAIN_DIR
    tar xjvf $KERNEL_TOOLCHAIN_TARBALL
fi

export PATH=$KERNEL_TOOLCHAIN_DIR/bin:$PATH

##### framebuffer resolution #####
if [ "$EURASIA_DC_SIZE" = "720P" ]; then \
    export DC_SIZE_720P=y
elif [ "$EURASIA_DC_SIZE" = "768P" ]; then \
    export DC_SIZE_768P=y
elif [ "$EURASIA_DC_SIZE" = "900P" ]; then \
    export DC_SIZE_900P=y
elif [ "$EURASIA_DC_SIZE" = "1080P" ]; then \
    export DC_SIZE_1080P=y
fi

##### Mali configs #####
export CONFIG_MALI_FB=y
export CONFIG_MALI_MIDGARD=m
export CONFIG_MODULES=y
export CONFIG_MALI_MIDGARD_DVFS=y
export KBUILD_EXTMOD=y


##### Build KM #####
INSTALL_FOLDER=$DDK_KM_DIR/binary2_"$EURASIA_3D_CHIP_NAME"_gpu_"$BUILD_MODE"
PROFILE="kernel-"$BUILD_MODE".config"

if [ $COMMAND = "clean" ]
then
    if [[ $K_TOOLCHAIN_NAME = *llvm* ]]; then
        make CC=clang clean -C $DDK_KM_DIR/kernel/drivers/base/mali_fb
        make CC=clang clean -C $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard
        rm -rf $INSTALL_FOLDER/
    else
        make clean -C $DDK_KM_DIR/kernel/drivers/base/mali_fb
        make clean -C $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard
        rm -rf $INSTALL_FOLDER/
    fi
    exit 0
elif [ $COMMAND = "make" ]
then
    if [ ! -d $INSTALL_FOLDER ]
    then
        mkdir -p $INSTALL_FOLDER
    fi
    export BUILD=$BUILD_MODE
    
    if [[ $K_TOOLCHAIN_NAME = *llvm* ]]; then
       make CC=clang -C $DDK_KM_DIR/kernel/drivers/base/mali_fb
       cp $DDK_KM_DIR/kernel/drivers/base/mali_fb/mali_fb.ko $INSTALL_FOLDER/
       make CC=clang -C $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard
       cp $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard/mali_kbase.ko $INSTALL_FOLDER/
    else
       make -C $DDK_KM_DIR/kernel/drivers/base/mali_fb
       cp $DDK_KM_DIR/kernel/drivers/base/mali_fb/mali_fb.ko $INSTALL_FOLDER/
       make -C $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard
       cp $DDK_KM_DIR/kernel/drivers/gpu/arm/midgard/mali_kbase.ko $INSTALL_FOLDER/
    fi
    exit 0
elif [ $COMMAND = "install" ]
then
    if [ ! -d $DISCIMAGE/vendor/lib/modules ]
    then
        mkdir -p $DISCIMAGE/vendor/lib/modules
    fi
    cp $DDK_KM_DIR/*.ko $DISCIMAGE/vendor/lib/modules
    echo "Install completed!"
    exit 0
elif [ $COMMAND = "sign" ]
then
    mkdir -p $INSTALL_FOLDER/signed
    if [ -f $INSTALL_FOLDER/mali_fb.ko ]
    then
        echo "signing mali_fb.ko"
        $KDIR/scripts/sign-file sha256 $KDIR/certs/rtk_signing_key.pem $KDIR/certs/signing_key.x509 $INSTALL_FOLDER/mali_fb.ko $INSTALL_FOLDER/signed/mali_fb.ko
    fi
    if [ -f $INSTALL_FOLDER/mali_kbase.ko ]
    then
        echo "signing mali_kbase.ko"
        $KDIR/scripts/sign-file sha256 $KDIR/certs/rtk_signing_key.pem $KDIR/certs/signing_key.x509 $INSTALL_FOLDER/mali_kbase.ko $INSTALL_FOLDER/signed/mali_kbase.ko
    fi
    exit 0
fi
